/* Name: main.c
 * Project: USB IR Remote Receiver with V-USB and Irmp
 * Author: Portisch
 * Creation Date: 15.03.2010
 * Copyright: (c) 2010 by Portisch
 * License: GNU GPL v2 (see License.txt), GNU GPL v3 or proprietary (CommercialLicense.txt)
 */

/*
This example should run on most AVRs with only little changes. No special
hardware resources except INT0 are used. You may have to change usbconfig.h for
different I/O pins for USB. Please note that USB D+ must be the INT0 pin, or
at least be connected to INT0 as well.
*/

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>  /* for sei() */
#include <util/delay.h>     /* for _delay_ms() */
#include <avr/eeprom.h>

#include "irmp.h"

#include <avr/pgmspace.h>   /* required by usbdrv.h */
#include "usbdrv.h"
#include "oddebug.h"        /* This is also an example for using debug macros */

#ifndef F_CPU
#error F_CPU unkown
#endif

/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

PROGMEM char usbHidReportDescriptor[122] = {
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x0b, 0x01, 0x00, 0x00, 0xff,  // USAGE (Vendor Defined Page 1:Vendor Usage 1)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x85, 0x01,                    //   REPORT_ID (1)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x06,                    //   REPORT_COUNT (6)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0x85, 0x02,                    //   REPORT_ID (2)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0x85, 0x03,                    //   REPORT_ID (3)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x06,                    //   REPORT_COUNT (6)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0x85, 0x04,                    //   REPORT_ID (4)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0x85, 0x05,                    //   REPORT_ID (5)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x06,                    //   REPORT_COUNT (6)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0x85, 0x06,                    //   REPORT_ID (6)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x02,                    //   REPORT_COUNT (2)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0x85, 0x07,                    //   REPORT_ID (7)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x04,                    //   REPORT_COUNT (4)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0xc0                           // END_COLLECTION
};

const char irmpVersion[4] = {
	0x0E,						//14.
	0x03,						//March
	0xDA, 0x07					//2010
};




/* The following variables store the status of the current data transfer */
static uchar    currentAddress   = 0;
static uchar    bytesRemaining   = 0;
static uchar	PowerOnEnabled   = 1;
static uchar	NewIRCodeAvailable 	= 0;

static uchar	* AdressPowerOn			= 0;
static uchar	* AdressTrainedCode 	= 1;

static uchar    replyBuf[7];

// startup delay is needed for USB Init
// first run timer0, after a few seconds timer1 is starting
// and IR polling starts
static uchar	StartupDelay = 0xFF;

static IRMP_DATA irmp_data;
static IRMP_DATA last_received_irmp_data;
static IRMP_DATA trained_irmp_data;


/* ------------------------------------------------------------------------- */
/* main functions for irmp
 */

//define output pin for power on switch:
#define SWITCH_PORT 		PORTC          /* PORTx - register for Switch output */
#define SWITCH_BIT  		PC5            /* bit for button input/output */
#define SWITCH_DDR  		DDRC           /* Switch data direction register */


void
timer_init (void)
{

	/* Timer 0 init  */
	// Timer0 Counter to zero 
	TCNT0 = 0x00;   
	// Prescaler 1024 / Clock <- CPU 
	TCCR0 = ( 1 << CS02 | 0 << CS01 | 1 << CS00 );
	
	#if defined (__AVR_ATmega8__) || defined (__AVR_ATmega16__) || defined (__AVR_ATmega32__) || defined (__AVR_ATmega64__) || defined (__AVR_ATmega162__)
		//Timer0 Start
		TIMSK = 1 << TOIE0;   
	#else
		//Timer0 Start
	 	TIMSK1  = 1 << TOIE0;                                                    // OCIE1A: Interrupt by timer compare
	#endif	// __AVR...

	// may adjust IR polling rate here to optimize IR receiving:
	OCR1A   =  (F_CPU / F_INTERRUPTS) - 1;                                    // compare value: 1/10000 of CPU frequency
	//OCR1A   =  0x550;                                    // compare value: 1/10000 of CPU frequency
}

void 
init_io(void) 
{
        /* set switch pin as digital output */
        SWITCH_DDR = _BV (SWITCH_BIT); 
}


ISR(TIMER0_OVF_vect) /* Interrupt Routine */ 
{
	//dec startup delay var
	StartupDelay--;  

	//reset Timer0 counter
   	TCNT0 = 0x00;
   
   	//startup delay finished?
	if (StartupDelay == 0   )
	{
		//stop Timer0
		TCCR0 = ( 0 << CS02 | 0 << CS01 | 0 << CS00 );

		//config IR polling timer
		// switch CTC Mode on, set prescaler to 1
		TCCR1B  = (1 << WGM12) | (1 << CS10);

		//disable Timer0, enable Timer1 for IR polling
		#if defined (__AVR_ATmega8__) || defined (__AVR_ATmega16__) || defined (__AVR_ATmega32__) || defined (__AVR_ATmega64__) || defined (__AVR_ATmega162__)
			// Timer0 ISR deaktivate, Timer1A ISR activate 
			TIMSK = ( 0 << TOIE0 | 1 << OCIE1A );   		
		#else
		 	// Timer0 ISR deaktivate, Timer1A ISR activate 
			TIMSK1  = ( 0 << TOIE0 | 1 << OCIE1A );                                                    // OCIE1A: Interrupt by timer compare
		#endif	// __AVR...
	}
}  

 

/*---------------------------------------------------------------------------------------------------------------------------------------------------
 * timer 1 compare handler, called every 1/10000 sec
 *---------------------------------------------------------------------------------------------------------------------------------------------------
 */
// Timer 1 output compare A interrupt service routine

ISR(TIMER1_COMPA_vect)
{

  irmp_ISR();                                                           // call irmp ISR
  // call other timer interrupt routines...
}

/* ------------------------------------------------------------------------- */
/* usbFunctionWrite() is called when the host sends a chunk of data to the
 * device. For more information see the documentation in usbdrv/usbdrv.h.
 */
uchar   usbFunctionWrite(uchar *data, uchar len)
{
static uchar i;
static uchar    Buf[8]; 

    if(bytesRemaining == 0)
        return 1;               /* end of transfer */
    if(len > bytesRemaining)
        len = bytesRemaining;

	//copy data by ignoring the first byte (reportnumber)
	for (i=0; i<len; i++){
		Buf[i] = data[i+1];
	};


	if ( len == sizeof(PowerOnEnabled))
	{
		// store new value of PowerOn in eeprom
		eeprom_write_block(&Buf, (uchar *)0 + currentAddress, len);

		//update PowerOn flag:
		memcpy(&PowerOnEnabled, &Buf, sizeof(PowerOnEnabled));
	}
	else if ( len == sizeof(irmp_data))
	{
		// store new trained IR code in eeprom
		eeprom_write_block(&Buf, (uchar *)0 + currentAddress, len);

		//update trained IR code:
		memcpy(&trained_irmp_data, &Buf, sizeof(irmp_data));
	} 
	else if ( len == sizeof(OCR1A))
	{
		// update new IR polling time
		OCR1A = (Buf[1] << 8 ) | Buf[0];
	} 
    currentAddress += len;
    bytesRemaining -= len;
    return bytesRemaining == 0; /* return 1 if this was the last chunk */
}


/* ------------------------------------------------------------------------- */

uchar   usbFunctionSetup(uchar data[8]) 
{ 

usbRequest_t    *rq = (void *)data;


    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    // HID class request 
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  // wValue: ReportType (highbyte), ReportID (lowbyte)

			usbMsgPtr = replyBuf;

            if(rq->wValue.bytes[0] == 1){   /* ReportID 1 */ 

				if ( NewIRCodeAvailable == 1 )
				{
					//copy report id
					memcpy(&replyBuf[0], &rq->wValue.bytes[0], sizeof(uchar));

					// copy last received IR code to reply buffer
					memcpy(&replyBuf[1], &last_received_irmp_data, sizeof(irmp_data));

					// clear new IR code flag
					NewIRCodeAvailable = 0;											 
                	return 7;
				}
				else
				{
					//copy report id
					memcpy(&replyBuf[0], &rq->wValue.bytes[0], sizeof(uchar));

					// clear new IR code flag
					NewIRCodeAvailable = 0;
					return 1;
				}
								
            }else if(rq->wValue.bytes[0] == 2){ 		/* ReportID 2 */ 

                //copy report id
				memcpy(&replyBuf[0], &rq->wValue.bytes[0], sizeof(uchar));

				// copy PowerOn flag to reply buffer
				eeprom_read_block(&replyBuf[1], AdressPowerOn, sizeof(PowerOnEnabled)); 
                return 2;  
            }else if(rq->wValue.bytes[0] == 3){ 		/* ReportID 3 */ 

				//copy report id
				memcpy(&replyBuf[0], &rq->wValue.bytes[0], sizeof(uchar));

				// copy trained IR code to reply buffer
				eeprom_read_block(&replyBuf[1], AdressTrainedCode, sizeof(irmp_data));  
                return 7;  
            }else if(rq->wValue.bytes[0] == 7){ 		/* ReportID 7 */ 

				//copy report id
				memcpy(&replyBuf[0], &rq->wValue.bytes[0], sizeof(uchar));

				// copy trained IR code to reply buffer
				memcpy(&replyBuf[1], &irmpVersion, sizeof(irmpVersion));
                return 5;  
            }
			 
        }else if(rq->bRequest == USBRQ_HID_SET_REPORT){ 
            if(rq->wValue.bytes[0] == 4){   			/* ReportID 4 */
				//store PowerOnEnabled in EEProm:
	            bytesRemaining = sizeof(PowerOnEnabled);
	            currentAddress = AdressPowerOn;
	            return USB_NO_MSG;  // use usbFunctionWrite() to receive data from host	
            }else if(rq->wValue.bytes[0] == 5){  		/* ReportID 5 */
				//store IR Code in EEProm
	            bytesRemaining = sizeof(irmp_data);
	            currentAddress = AdressTrainedCode;
	            return USB_NO_MSG;  // use usbFunctionWrite() to receive data from host	    
            }else if(rq->wValue.bytes[0] == 6){  		/* ReportID 6 */
				//update new IR polling time
	            bytesRemaining = sizeof(OCR1A);
	            return USB_NO_MSG;  // use usbFunctionWrite() to receive data from host	
            }   			         
        } 
    }else{ 
        // ignore vendor type requests, we don't use any
    } 
    return 0; 
} 

/* ------------------------------------------------------------------------- */

int main(void)
{
uchar   i;

    wdt_enable(WDTO_1S);
    /* Even if you don't use the watchdog, turn it off here. On newer devices,
     * the status of the watchdog (on/off, period) is PRESERVED OVER RESET!
     */
    /* RESET status: all port bits are inputs without pull-up.
     * That's the way we need D+ and D-. Therefore we don't need any
     * additional hardware initialization.
     */
    odDebugInit();
    usbInit();

	init_io();

	//clear irmp_data
	memset(&irmp_data, 0x00, sizeof(irmp_data));  

	//clear last_received_irmp_data
	memset(&last_received_irmp_data, 0x00, sizeof(irmp_data));  
	
	//clear trained_irmp_data
	memset(&trained_irmp_data, 0x00, sizeof(irmp_data));


	irmp_init();                                                              // initialize irmp code
	timer_init();                                                             // initialize timer

    usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
    i = 0;
    while(--i){             /* fake USB disconnect for > 250 ms */
        wdt_reset();
        _delay_ms(1);
    }
    usbDeviceConnect();
	//global int
    sei();

	//set PowerOn Enabled to default:
	eeprom_write_block(&PowerOnEnabled, AdressPowerOn, sizeof(PowerOnEnabled));

	//read last saved trained IR code in eeprom
	eeprom_read_block(&trained_irmp_data, AdressTrainedCode, sizeof(irmp_data));

    for(;;){                /* main event loop */
        wdt_reset();

		//check if new IR code is aviable
	    if (irmp_get_data (&irmp_data))
	    {
			//do a USB poll
			usbPoll();
	        // ir signal decoded, do something here...
	        // irmp_data.protocol is the protocol, see irmp.h
	        // irmp_data.address is the address/manufacturer code of ir sender
	        // irmp_data.command is the command code

			// save received irmp data
			memcpy(&last_received_irmp_data, &irmp_data, sizeof(irmp_data)) ;
			
			// check if code is already trained:
			if ((trained_irmp_data.protocol != 0xFF) & (trained_irmp_data.protocol != 0x00))
			{

				// IR code trained,
				// if PowerOn function is enabled check it
				if (PowerOnEnabled == 1)
				{
					//compare trained code with last received code
					if (memcmp(&last_received_irmp_data, &trained_irmp_data, sizeof(irmp_data) - 1) == 0)				 
					{	
						//siwtch on output port pin for ~250ms
						SWITCH_PORT ^= _BV(SWITCH_BIT);
					    while(--i)
						{
					        wdt_reset();
							usbPoll();
					        _delay_us(500);
					    }
						//switch off output port pin
						SWITCH_PORT ^= _BV(SWITCH_BIT);
					}
					else
					{
						// set flag for received new IR code
						NewIRCodeAvailable = 1;
					}
				}
				else
				{
					// set flag for received new IR code
					NewIRCodeAvailable = 1;
				}
			}
			else
			{
				//no IR code trained, save it:
				eeprom_write_block(&last_received_irmp_data, AdressTrainedCode, sizeof(irmp_data));
				//update trained_irmp_data
				memcpy(&trained_irmp_data, &last_received_irmp_data, sizeof(irmp_data));
			}
	    }
			//do a USB poll
        	usbPoll();
    }
    return 0;
}

/* ------------------------------------------------------------------------- */
